/*
 * utility.hpp
 *
 * This file is part of Algorithms library toolkit.
 * Copyright (C) 2017 Jan Travnicek (jan.travnicek@fit.cvut.cz)

 * Algorithms library toolkit is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * Algorithms library toolkit is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with Algorithms library toolkit.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Created on: Apr 7, 2016
 * Author: Jan Travnicek
 */

#ifndef __UTILITY_HPP_
#define __UTILITY_HPP_

#include <utility>
#include <cstddef>

#include "type_traits.hpp"

namespace ext {

/**
 * \brief
 * Extension of sizeof mimicking behavior of regular sizeof operator.
 *
 * Provides field size as result of sizeof operator
 *
 * \tparam T the tested type
 */
template<class T>
struct SizeOf {
	/**
	 * \brief
	 * The value of sizeof operator on the type T.
	 */
	static const size_t size = sizeof(T);
};

/**
 * \brief
 * Specialised sizeof extension for void type. The void type is resulting into the size 1.
 *
 * Provides field size as result of sizeof operator
 */
template<>
struct SizeOf<void> {
	/**
	 * \brief
	 * The size for void type is set to 1.
	 */
	static const size_t size = 1;
};

/**
 * \brief
 * Extension of alignof mimicking behavior of regular alignof operator.
 *
 * Provides field align as result of sizeof operator
 *
 * \tparam T the tested type
 */
template<class T>
struct AlignOf {
	/**
	 * \brief
	 * The value of alignof operator on the type T.
	 */
	static const size_t align = alignof(T);
};

/**
 * \brief
 * Specialised alignof extension for void type. The void type is resulting into the align 1.
 *
 * Provides field align as result of alignof operator
 */
template<>
struct AlignOf<void> {
	/**
	 * \brief
	 * The align for void type is set to 1.
	 */
	static const size_t align = 1;
};

/**
 * \brief
 * Class representing wrapper of dynamically allocated object behaving like rvalue reference.
 *
 * The class takes ownership of the dynamically allocated object
 *
 * \tparam the type of value wrapped
 */
template < class T >
class rvalue_ref {
	/**
	 * \brief
	 * Owned dynamically allocated object
	 */
	T * holder;

public:
	/**
	 * \brief
	 * Constructor of the rvalue_ref from dynaically allocated object. The class will take ownerhip of the pointer.
	 */
	explicit rvalue_ref ( T * ptr ) : holder ( ptr ) {

	}

	/**
	 * \brief
	 * User conversion constructor from rvalue_ref for type U where T is its predecessor
	 *
	 * \tparam U the subtype of T
	 *
	 * \param other the other instance
	 */
	template < class U >
	rvalue_ref ( rvalue_ref < U > && other ) : holder ( other.holder ) {
		other.holder = nullptr;
	}

	/**
	 * \brief
	 * Copy construction is not allowed.
	 */
	rvalue_ref ( const rvalue_ref & ) = delete;

	/**
	 * \brief
	 * Move construction will transfer ownership of the pointer from source instance to newly created one.
	 *
	 * \param other the other instance
	 */
	rvalue_ref ( rvalue_ref && other) : holder( other.holder ) {
		other.holder = nullptr;
	};

	/**
	 * \brief
	 * Copy assignment is not allowed.
	 */
	rvalue_ref & operator = ( const rvalue_ref & ) = delete;

	/**
	 * \param
	 * Move assignment will swap holded pointers between the two instances.
	 *
	 * \param other the other instance.
	 */
	rvalue_ref & operator = ( rvalue_ref && other ) {
		using std::swap;

		swap(holder, other.holder);

		return *this;
	}

	/**
	 * \brief
	 * Destructor will free the owned instance.
	 */
	~rvalue_ref ( ) {
		delete holder;
	}

	/**
	 * \brief
	 * Arrow operator to chain dereference.
	 *
	 * \return the owned pointer
	 */
	T * operator ->( ) && {
		return holder;
	}

	/**
	 * \brief
	 * Arrow operator to chain dereference.
	 *
	 * \return the owned pointer
	 */
	T * operator ->( ) & {
		return holder;
	}

	/**
	 * \brief
	 * Arrow operator to chain dereference.
	 *
	 * \return the owned pointer
	 */
	const T * operator ->( ) const & {
		return holder;
	}

	/**
	 * \brief
	 * Allow automatic cast to rvalue reference.
	 *
	 * \return the holded value
	 */
	operator T && ( ) {
		return std::move ( * holder );
	}

	/**
	 * \brief
	 * Getter of the holded value.
	 *
	 * \return constant reference to the holded value
	 */
	const T & get ( ) const {
		return * holder;
	}

	template < class U >
	friend class rvalue_ref;
};

/**
 * \brief
 * Allow moving of copied instance of the source. Designed for copyconstructible objects.
 *
 * \tparam T the type of processed instance
 *
 * \param param the source instance
 *
 * \return copy to be moved from
 */
template < class T, typename std::enable_if < ! ext::has_clone < T >::value >::type * = nullptr >
inline T move_copy ( const T & param ) {
	return T ( param );
}

/**
 * \brief
 * Allow moving of copied instance of the source. Designed for cloneable objects.
 *
 * \tparam T the type of processed instance
 *
 * \param param the source instance
 *
 * \return wrapper of dynamically allocated copy supporting cast to rvalue reference to T
 */
template < class T, typename std::enable_if < ext::has_clone < T >::value >::type * = nullptr >
inline rvalue_ref < T > move_copy ( const T & param ) {
	return rvalue_ref < T > { param.clone ( ) };
}

 // TODO remove after switching to c++17 (as_const)
template < class T >
inline const T & make_const ( T & ptr ) {
	return static_cast < const T & > ( ptr );
}

template < class T >
inline const T & make_const ( const T & ptr ) {
	return ptr;
}

} /* namespace ext */

#endif /* __UTILITY_HPP_ */
